/** @file
 * @brief Config class implementation. This class represents the actual configuration.
 *
 *
 *
 * @copyright Intra2net AG
 * @license GPLv2
*/

#include "config.h"

#include "serviceholder.h"
#include "dhs.h"
#include "ods.h"

#include <time.h>
#include <iostream>
#include <fstream>

#include <boost/foreach.hpp>
#include <boost/filesystem.hpp>
#include <boost/regex.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/archive_exception.hpp>
#include <boost/serialization/export.hpp>
#include <boost/serialization/shared_ptr.hpp>
#include <boost/algorithm/string.hpp>

// Following boost macros are needed for serialization of derived classes through a base class pointer (Service *).
BOOST_CLASS_EXPORT_GUID(ODS, "ODS")
BOOST_CLASS_EXPORT_GUID(DHS, "DHS")

namespace po = boost::program_options;
namespace fs = boost::filesystem;
namespace ba = boost::algorithm;

using namespace std;

/**
 * Default Constructor. Available command line and config file options with their default values are defined here.
 */
Config::Config(Logger::Ptr _log)
    : Log(new Logger)
    , DaemonMode(false)
    , Loglevel(0)
    , Syslog(false)
    , ConfigPath("/etc/bpdyndnsd")
    , EnableIPv6(false)
    , WebcheckIpUrl("")
    , WebcheckIpUrlAlt("")
{
    // initialize Logger
    Log = _log;

    // Available service description config options
    po::options_description opt_desc_service("Service description options");
    opt_desc_service.add_options()
        ("protocol",po::value<string>(),"The service protocol.")
        ("host",po::value<string>(),"The hostname to update.")
        ("login",po::value<string>(),"Login name.")
        ("password",po::value<string>(),"Corresponding password.")
        ("update_interval",po::value<int>()->default_value(-1),"Update interval in minutes.")
        ("max_updates_within_interval",po::value<int>()->default_value(-1),"How many updates can be made in one interval.")
    ;

    // Available command line only options
    po::options_description opt_desc_cmd_only("Command line only options");
    opt_desc_cmd_only.add_options()
        ("help,?","Show help.")
        ("version,v","Show version.")
        ("config,c",po::value<string>()->default_value("/etc/bpdyndnsd"),"Set the config path.")
    ;

     // Available generic options. Valid on cmd or in config file.
    po::options_description opt_desc_generic("Generic config options");
    opt_desc_generic.add_options()
        ("daemon_mode",po::value<bool>()->default_value(false),"Run as system daemon.")
        ("loglevel",po::value<int>()->default_value(0),"Loglevel.")
        ("syslog",po::value<bool>()->default_value(false),"Use syslog facility.")
        ("enable_ipv6",po::value<bool>()->default_value(false),"Try to use IPv6.")
        ("webcheck_url",po::value<string>()->default_value(""),"Use this URL to determine IP.")
        ("webcheck_url_alt",po::value<string>()->default_value(""),"Use this alternative URL to determine IP.")
    ;

    // Define valid command line parameters
    Options_descriptionPtr _opt_desc_cmd(new po::options_description("Command line options"));
    OptDescCmd = _opt_desc_cmd;
    _opt_desc_cmd.reset();
    OptDescCmd->add(opt_desc_cmd_only);
    OptDescCmd->add(opt_desc_generic);
    OptDescCmd->add(opt_desc_service);

    // Define valid config file options
    Options_descriptionPtr _opt_desc_conf_main(new po::options_description("Config file options"));
    OptDescConfMain = _opt_desc_conf_main;
    _opt_desc_conf_main.reset();
    OptDescConfMain->add(opt_desc_generic);

    // Define valid service file options
    Options_descriptionPtr _opt_desc_conf_service(new po::options_description("Service file options"));
    OptDescConfService = _opt_desc_conf_service;
    _opt_desc_conf_service.reset();
    OptDescConfService->add(opt_desc_service);
}


/**
 * Default Destructor
 */
Config::~Config()
{
}


/**
 * This function serializes all Service objects in Services and OldServices (where the update Timeout isn't expired) into a specified file.
 * @return 0 if all is fine, -1 if output file could not be opened for reading or error while serializing.
 */
int Config::serialize_services()
{
    // Put Services and OldServices into Serviceholder.
    Serviceholder::Ptr service_holder(new Serviceholder());

    BOOST_FOREACH(Service::Ptr &service, Services)
    {
        service_holder->add_service(service);
    }

    int current_time = time(NULL);

    BOOST_FOREACH(Service::Ptr &service, OldServices)
    {
        if ( ( service->get_last_updates()->front() + (service->get_update_interval()*60) ) >= current_time )  // UpdateInterval timeout of service isn't expired.
            service_holder->add_service(service);
    }

    // Serialize Serviceholder into file.
    ofstream ofs(OBJECT_FILE);
    if ( ofs.is_open() )
    {
        Serviceholder* _service_holder = service_holder.get();
        try
        {
            boost::archive::text_oarchive oa(ofs);
            oa << _service_holder;
        }
        catch( boost::archive::archive_exception e )
        {
            Log->print_exception_serialize(e.what());
            ofs.close();
            return -1;
        }

        ofs.close();
    }
    else
    {
        Log->print_error_opening_rw(OBJECT_FILE);
        return -1;
    }

    Log->print_serialized_objects_success();

    return 0;
}


/**
 * This function de-serializes the Serviceholder (containing Services) from the object file.
 * @return 0 if all is fine, -1 if object file couldn't be opened for reading or error while de-serializing.
 */
int Config::deserialize_services()
{
    // test if OBJECT_FILE exists
    fs::path object_file = fs::system_complete(fs::path(OBJECT_FILE));
    if ( !fs::exists(object_file) )
    {
        // There is no object file, possibly first program start, continue without recovering old Services' state.
        Log->print_no_object_file(OBJECT_FILE);
        return 0;
    }

    ifstream ifs(OBJECT_FILE);
    if ( ifs.is_open() )
    {
        // deserialize Serviceholder
        Serviceholder* _service_holder;
        try
        {
            boost::archive::text_iarchive ia(ifs);
            ia >> _service_holder;       // ends up in default constructor call for Serviceholder
        }
        catch( boost::archive::archive_exception e )
        {
            Log->print_exception_deserialize(e.what());
            ifs.close();
            return -1;
        }
        Serviceholder::Ptr service_holder(_service_holder);
        ifs.close();

        // Get the list of old Services from de-serialized Serviceholder object.
        list<Service::Ptr> _old_services = service_holder->get_hold_services();

        // Put all de-serialized Services into OldServices member and
        // compare new Services (generated from config file and cmd) with old Services (de-serialized from object file)
        // if identical Service was found, adopt the value from old Service to recover Services' state.
        BOOST_FOREACH(Service::Ptr &old_service, _old_services)
        {
            OldServices.push_back(old_service);
            Log->print_service_object("Deserialized following Service object:", old_service->get_protocol(), old_service->get_hostname(), old_service->get_login() ,old_service->get_password(), old_service->get_update_interval(), old_service->get_max_updates_within_interval(), old_service->get_actual_ip(), old_service->get_last_updates());
            BOOST_FOREACH(Service::Ptr &service, Services)
            {
                if ( *service == *old_service )
                {
                    service->set_last_updates(old_service->get_last_updates());
                    service->set_actual_ip(old_service->get_actual_ip());
                    Log->print_service_object("New Service object with adopted values:", service->get_protocol(), service->get_hostname(), service->get_login() ,service->get_password(), service->get_update_interval(), service->get_max_updates_within_interval(), service->get_actual_ip(), service->get_last_updates());
                    // We have adopted the values of the old_service. Just set lastupdated and timeout to 0, so this old_service wont be serialized.
                    old_service->set_update_interval(0);
                }
            }
        }
        Log->print_deserialized_objects_success();
    }
    else
    {
        Log->print_error_opening_r(OBJECT_FILE);
        return -1;
    }

    return 0;
}


/**
 * Parses the command line arguments and does the needed actions.
 * @param argc Command line argument number given to main.
 * @param argv[] Pointer to command line argument array given to main.
 * @return 0 if all is fine, -1 if not.
 */
int Config::parse_cmd_line(int argc, char *argv[])
{
    try
    {
        po::store(po::parse_command_line(argc, argv, *this->OptDescCmd), VariablesMap);
        po::notify(VariablesMap);

        if ( VariablesMap.count("help") )
        {
            Log->print_usage(OptDescCmd);
            return -1;
        }
        else if ( VariablesMap.count("version") )
        {
            Log->print_version();
            return -1;
        }

        // Create a service object if all needed options are set on the command line
        if ( VariablesMap.count("protocol") && VariablesMap.count("host") && VariablesMap.count("login") && VariablesMap.count("password") )
        {
            // Get the cmd parameter values for protocol host login and password
            string protocol = VariablesMap["protocol"].as<string>();
            string host = VariablesMap["host"].as<string>();
            string login = VariablesMap["login"].as<string>();
            string password = VariablesMap["password"].as<string>();

            protocol = ba::to_lower_copy(protocol);

            int update_interval = 0;
            if ( VariablesMap.count("update_interval") )
                update_interval = VariablesMap["update_interval"].as<int>();

            int max_updates_within_interval = 0;
            if ( VariablesMap.count("max_updates_within_interval") )
                max_updates_within_interval = VariablesMap["max_updates_within_interval"].as<int>();

            Service::Ptr service = create_service(protocol,host,login,password,update_interval,max_updates_within_interval);
            if ( service )
                Services.push_back(service);
            else
                return -1;
        }
        else if ( VariablesMap.count("protocol") || VariablesMap.count("host") || VariablesMap.count("login") || VariablesMap.count("password") )
        {
            Log->print_missing_cmd_service_option();
            Log->print_usage(OptDescCmd);
            return -1;
        }

        if ( VariablesMap.count("config") )
        {
            fs::path full_config_path = fs::system_complete(fs::path(VariablesMap["config"].as<string>()));
            ConfigPath = full_config_path.string();
            if ( !fs::exists(full_config_path) ||  !fs::is_directory(full_config_path) )
            {
                // Config path doesn't exist or is not a directory
                Log->print_error_config_path(ConfigPath);
                return -1;
            }
        }

        if ( VariablesMap.count("daemon_mode") )
            DaemonMode = VariablesMap["daemon_mode"].as<bool>();

        if ( VariablesMap.count("loglevel") )
            Loglevel = VariablesMap["loglevel"].as<int>();

        if ( VariablesMap.count("syslog") )
            Syslog = VariablesMap["syslog"].as<bool>();

        if ( VariablesMap.count("enable_ipv6") )
            EnableIPv6 = VariablesMap["enable_ipv6"].as<bool>();

        if ( VariablesMap.count("webcheck_url") )
            WebcheckIpUrl = VariablesMap["webcheck_url"].as<string>();

        if ( VariablesMap.count("webcheck_url_alt") )
            WebcheckIpUrlAlt = VariablesMap["webcheck_url_alt"].as<string>();

    }
    catch(po::unknown_option e)
    {
        Log->print_unknown_cmd_option(e.what());
        Log->print_usage(OptDescCmd);
        return -1;
    }
    catch(po::multiple_occurrences e)
    {
        Log->print_multiple_cmd_option(e.what());
        Log->print_usage(OptDescCmd);
        return -1;
    }
    return 0;
}


/**
 * Creates a Service object from the given parameters.
 * @param protocol Protocol to use.
 * @param host Hostname to update.
 * @param login Login.
 * @param password Password.
 * @return A pointer to the created Service object.
 */
Service::Ptr Config::create_service(const string &protocol,const string &hostname, const string &login, const string &password, const int update_interval, const int max_updates_within_interval)
{
    if(protocol == "dhs")
    {
        Service::Ptr service_dhs(new DHS(protocol,hostname,login,password,Log,update_interval,max_updates_within_interval));
        return service_dhs;
    }
    else if(protocol == "ods")
    {
        Service::Ptr service_ods(new ODS(protocol,hostname,login,password,Log,update_interval,max_updates_within_interval));
        return service_ods;
    }
    else
    {
        Log->print_unknown_protocol(protocol);
        Service::Ptr service;
        return service;
    }
}


/**
 * Loads a service config file, invoked by load_config_from_files.
 * @param full_filename Filename of the service config file to load.
 * @return 0 if all is fine, -1 otherwise.
 */
int Config::load_service_config_file(const string& full_filename)
{
    Log->print_load_service_conf(full_filename);

    ifstream service_config_file(full_filename.c_str(),ifstream::in);
    if(service_config_file.is_open())
    {
        try
        {
            po::variables_map vm;
            po::parsed_options parsed_service_options = po::parse_config_file(service_config_file,*this->OptDescConfService,true);
            po::store(parsed_service_options,vm);
            po::notify(vm);

            if(vm.count("protocol") && vm.count("host") && vm.count("login") && vm.count("password"))
            {
                // create the corresponding service
                string protocol = vm["protocol"].as<string>();
                string host = vm["host"].as<string>();
                string login = vm["login"].as<string>();
                string password = vm["password"].as<string>();

                protocol = ba::to_lower_copy(protocol);

                int update_interval = 0;
                if ( VariablesMap.count("update_interval") )
                    update_interval = VariablesMap["update_interval"].as<int>();

                int max_updates_within_interval = 0;
                if ( VariablesMap.count("max_updates_within_interval") )
                    max_updates_within_interval = VariablesMap["max_updates_within_interval"].as<int>();

                Service::Ptr service = create_service(protocol,host,login,password,update_interval,max_updates_within_interval);
                if ( service )
                {
                    Services.push_back(service);
                    Log->print_service_object("New Service object from config:", service->get_protocol(), service->get_hostname(), service->get_login() ,service->get_password(), service->get_update_interval(), service->get_max_updates_within_interval(), service->get_actual_ip(), service->get_last_updates());
                }
            }
        }
        catch ( po::unknown_option e )
        {
            // unknown option in config file detected
            service_config_file.close();
            Log->print_unknown_service_conf_option(e.what());
            return -1;
        }
        service_config_file.close();
    }
    else
    {
        // error opening service config file for reading
        Log->print_error_opening_r(full_filename);
        return -1;
    }
    return 0;
}


/**
 * Loads the main config file, invoked by load_config_from_files
 * @param full_filename The full filename of the main config file to load
 * @return 0 if all is fine, -1 otherwise
 */
int Config::load_main_config_file(const string& full_filename)
{
    Log->print_load_main_conf(full_filename);

    ifstream main_config_file(full_filename.c_str(),ifstream::in);
    if(main_config_file.is_open())
    {
        try
        {
            po::parsed_options parsed_main_options = po::parse_config_file(main_config_file,*this->OptDescConfMain,true);
            po::store(parsed_main_options,VariablesMap);
            po::notify(VariablesMap);

        if ( VariablesMap.count("daemon_mode") )
            DaemonMode = VariablesMap["daemon_mode"].as<bool>();

        if ( VariablesMap.count("loglevel") )
            Loglevel = VariablesMap["loglevel"].as<int>();

        if ( VariablesMap.count("syslog") )
            Syslog = VariablesMap["syslog"].as<bool>();

        if ( VariablesMap.count("enable_ipv6") )
            EnableIPv6 = VariablesMap["enable_ipv6"].as<bool>();

        if ( VariablesMap.count("webcheck_url") )
            WebcheckIpUrl = VariablesMap["webcheck_url"].as<string>();

        if ( VariablesMap.count("webcheck_url_alt") )
            WebcheckIpUrlAlt = VariablesMap["webcheck_url_alt"].as<string>();

        }
        catch ( po::unknown_option e )      // at the moment 04-08-2009 this exception is never thrown :-(
        {
            // unknown option in main config file detected
            main_config_file.close();
            Log->print_unknown_main_conf_option(e.what());
            return -1;
        }
        main_config_file.close();
    }
    else
    {
        // error opening main config file for reading
        Log->print_error_opening_r(full_filename);
        return -1;
    }
    return 0;
}


/**
 * Loads the main and the service config file and does the needed action.
 * @param config_path The path to the config directory.
 * @return 0 if all is fine, -1 otherwise
 */
int Config::load_config_from_files()
{
    fs::path full_config_path = fs::path(ConfigPath);

    fs::directory_iterator end_iter;
    for ( fs::directory_iterator dir_itr(full_config_path) ; dir_itr != end_iter ; ++dir_itr )
    {
        if( fs::is_regular_file( dir_itr->status() ) )
        {
            string actual_file = dir_itr->path().filename();
            boost::regex expr(".*\\.conf?");
             // If it is the main config file do the following
            if ( actual_file == "bpdyndnsd.conf" )
            {
                // Load the main config file
                string full_filename = dir_itr->path().string();
                if ( load_main_config_file(full_filename) != 0 )
                    return -1;
            }
            // If it is a service definition file *.conf, parse it and generate the corresponding service
            else if ( boost::regex_search( actual_file,expr ) )
            {
                string full_filename = dir_itr->path().string();
                if ( load_service_config_file(full_filename) != 0 )
                    return -1;
            }
        }
    }
    // Config file successfully loaded
    Log->print_conf_loaded(ConfigPath);
    return 0;
}


/**
 * Getter method for Service list member.
 * @return Pointer to a list of Service's.
 */
list<Service::Ptr> Config::get_services() const
{
    return this->Services;
}


/**
 * Getter method for member OptDescCmd.
 * @return options_description*.
 */
Config::Options_descriptionPtr Config::get_opt_desc_cmd() const
{
    return OptDescCmd;
}


/**
 * Getter method for member OptDescConfMain.
 * @return options_description*.
 */
Config::Options_descriptionPtr Config::get_opt_desc_conf_main() const
{
    return OptDescConfMain;
}


/**
 * Getter method for member OptDescConfService.
 * @return options_description*.
 */
Config::Options_descriptionPtr Config::get_opt_desc_conf_service() const
{
    return OptDescConfService;
}


/**
 * Getter for member Loglevel.
 * @return Member Loglevel.
 */
int Config::get_loglevel() const
{
    return Loglevel;
}


/**
 * Getter for member DaemonMode.
 * @return TRUE if enabled, FALSE if disabled.
 */
bool Config::get_daemon_mode() const
{
    return DaemonMode;
}


/**
 * Resets all shared Service pointers and clears the Services list.
 */
void Config::delete_services()
{
    BOOST_FOREACH( Service::Ptr &service, Services )
    {
        service.reset();
    }
    Services.clear();

    BOOST_FOREACH( Service::Ptr &service, OldServices )
    {
        service.reset();
    }
    OldServices.clear();
}


/**
 * Deletes the map with the previously parsed options.
 * This is needed in case we reload the config and don't want the old cmd options to overwrite new config file options.
 */
void Config::delete_variables_map()
{
    VariablesMap.clear();

    po::variables_map _variables_map;
    VariablesMap = _variables_map;
}


/**
 * Getter for member Syslog.
 * @return True if logging through syslog is enabled, false otherwise.
 */
bool Config::get_syslog() const
{
    return Syslog;
}


/**
 * Getter for member EnableIPv6
 * @return Wether IPv6 should be used or not.
 */
bool Config::get_enable_ipv6() const
{
    return EnableIPv6;
}


/**
 * Getter for member WebcheckIpUrl
 * @return The primary IP Webcheck URL
 */
string Config::get_webcheck_ip_url() const
{
    return WebcheckIpUrl;
}


/**
 * Getter for member WebcheckIpUrlAlt
 * @return The alternative IP Webcheck URL
 */
string Config::get_webcheck_ip_url_alt() const
{
    return WebcheckIpUrlAlt;
}
