/** @file
 * @brief Logger class implementation. This class implements the Logging facility.
 *
 *
 *
 * @copyright Intra2net AG
 * @license GPLv2
*/


#include "logger.h"

#include <syslog.h>
#include <sstream>

#include <boost/foreach.hpp>

namespace po = boost::program_options;

using namespace std;

/**
 * Default Constructor
 */
Logger::Logger()
    : Loglevel(0)
    , Syslog(false)
{
    set_log_facility(Loglevel,Syslog);
}


/**
 * Default Destructor
 */
Logger::~Logger()
{
}


/**
 * Decides if Logging through syslog if enabled or through std.
 * @param msg The message to log.
 */
void Logger::log_notice(const string& msg) const
{
    if ( Syslog )
        syslog(LOG_NOTICE,msg.c_str());
    else
        cout << msg << endl;;
}


/**
 * Decides if Logging through syslog if enabled or through std.
 * @param msg The message to log.
 */
void Logger::log_warning(const string& msg) const
{
    if ( Syslog )
        syslog(LOG_WARNING,msg.c_str());
    else
        cout << msg << endl;
}


/**
 * Decides if Logging through syslog if enabled or through std.
 * @param msg The message to log.
 */
void Logger::log_error(const string& msg) const
{
    if ( Syslog )
        syslog(LOG_ERR,msg.c_str());
    else
        cerr << msg << endl;
}


/**
 * Setter for member Loglevel.
 * @param _loglevel Value to set Loglevel to.
 */
void Logger::set_loglevel(const int _loglevel)
{
    Loglevel = _loglevel;
    cout << "Loglevel set" << endl;
}


/**
 * Getter for member Loglevel.
 * @return Loglevel.
 */
int Logger::get_loglevel() const
{
    return Loglevel;
}


/**
 * Setter for member Syslog.
 * @param _syslog Wether to log through syslog or not.
 */
void Logger::set_syslog(const bool _syslog)
{
    Syslog = _syslog;
}


/**
 * Getter for member Syslog.
 * @return True if logging through syslog is enabled, false otherwise.
 */
bool Logger::get_syslog() const
{
    return Syslog;
}


/**
 * Initialize the logging facility.
 */
void Logger::set_log_facility(const int _loglevel, const bool _syslog)
{
    Loglevel = _loglevel;
    Syslog = _syslog;

    if ( Syslog )
        openlog("bpdyndnsd",LOG_PID,LOG_DAEMON);
}


/**
 * Prints out the usage to the command line.
 */
void Logger::print_usage(const Options_descriptionPtr opt_desc) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Usage: bpdyndnsd [Command line options]" << "\n" << endl;
        msg << *opt_desc << endl;
        log_notice(msg.str());
    }
}


/**
 * Prints out the programm name and the given version string on stdout.
 * @param version Version string.
 */
void Logger::print_version() const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Bullet proof dynamic dns daemon.\nbpdyndnsd " << VERSION << "." << REVISION << "." << RELEASE << endl;
        log_notice(msg.str());
    }
}


/**
 * Prints out the successful parsing of the command line options.
 */
void Logger::print_cmd_parsed() const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "Command line options successfully parsed." << endl;
        log_notice(msg.str());
    }
}


void Logger::print_conf_files_parsed() const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "Config files successfully parsed." << endl;
        log_notice(msg.str());
    }
}


/**
 * Prints out the successful parsing of the config files.
 * @param config_path The specified config path.
 */
void Logger::print_conf_loaded(const string& config_path) const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "Config files successfully loaded in: " << config_path << endl;
        log_notice(msg.str());
    }
}


/**
 * Prints out the unsuccessful parsing of the config files.
 * @param config_path The specified config path.
 */
void Logger::print_conf_not_loaded(const string& config_path) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Config files couldn't be loaded in: " << config_path << endl;
        log_error(msg.str());
    }
}


/**
 * A file could not be opened for reading.
 * @param filename The filename.
 */
void Logger::print_error_opening_r(const string& filename) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Error opening file for reading: " << filename << endl;
        log_error(msg.str());
    }
}


/**
 * A file could not be opened for writing.
 * @param filename The filename.
 */
void Logger::print_error_opening_rw(const string& filename) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Error opening file for writing: " << filename << endl;
        log_error(msg.str());
    }
}


/**
 * Desctructor of specified class was called.
 * @param _class Name of the class.
 */
void Logger::print_destructor_call(const string& _class) const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "Destructor call: " << _class << endl;
        log_notice(msg.str());
    }
}


/**
 * Constructor of specified class was called.
 * @param _class Name of the class.
 */
void Logger::print_constructor_call(const string& _class) const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "Constructor call: " << _class << endl;
        log_notice(msg.str());
    }
}


/**
 * Update method for specified service was called.
 * @param service The service for which is the update running.
 */
void Logger::print_update_service(const string& service) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Running update for service: " << service << endl;
        log_notice(msg.str());
    }
}


/**
 * An unknown option on the command line was detected.
 * @param unknown_option The unknown option.
 */
void Logger::print_unknown_cmd_option(const string& unknown_option) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Unknown option on command line detected: " << unknown_option << endl;
        log_error(msg.str());
    }
}


/**
 * An unknown protocol was specified.
 * @param protocol The unknown protocol.
 */
void Logger::print_unknown_protocol(const string& protocol) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Unknown protocol defined: " << protocol << endl;
        log_error(msg.str());
    }
}


/**
 * Loading a service config file.
 * @param filename The service config file.
 */
void Logger::print_load_service_conf(const string& filename) const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "Loading service config file: " << filename << endl;
        log_notice(msg.str());
    }
}


/**
 * Loading the main config file.
 * @param filename The main config file.
 */
void Logger::print_load_main_conf(const string& filename) const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "Loading main config file: " << filename << endl;
        log_notice(msg.str());
    }
}


/**
 * There is an unknown option in a service config file.
 * @param unknown_option The unknown option.
 */
void Logger::print_unknown_service_conf_option(const string& unknown_option) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Unknown option in service config file detected: " << unknown_option << endl;
        log_error(msg.str());
    }
}


/**
 * There is an unknown option in the main config file.
 * @param unknown_option The unknown option.
 */
void Logger::print_unknown_main_conf_option(const string& unknown_option) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Unknown option in main config file detected: " << unknown_option << endl;
        log_error(msg.str());
    }
}


/**
 * The defined config path doesn't exist or is not a directory.
 * @param config_path The defined config path.
 */
void Logger::print_error_config_path(const string& config_path) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Config path doesn't exists or is not a diretory: " << config_path << endl;
        log_error(msg.str());
    }
}


/**
 * There is a missing command line option to initialize a service object.
 */
void Logger::print_missing_cmd_service_option() const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Missing option to initialize service. Protocol, host, login and password must be specified." << endl;
        log_error(msg.str());
    }
}


/**
 * Process running as daemon.
 * @param pid The pid of the daemon.
 */
void Logger::print_runnig_as_daemon(const int pid) const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "Runnig as daemon: " << pid << endl;
        log_notice(msg.str());
    }
}


/**
 * Prints out the daemon mode.
 * @param daemon_mode The daemon mode.
 */
void Logger::print_daemon_mode(const bool daemon_mode) const
{
    if ( 1 <= Loglevel )
    {
        string mode = "disabled";
        if (daemon_mode == true)
            mode = "enabled";
        ostringstream msg;
        msg << "Daemon mode is " << mode << "." << endl;
        log_notice(msg.str());
    }
}


/**
 * There was an error while trying to fork.
 */
void Logger::print_error_fork() const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Error while trying to fork." << endl;
        log_notice(msg.str());
    }
}


/**
 * A pid in the pidfile was found.
 * @param pid The pid found in the pidfile.
 */
void Logger::print_pid_found(const int pid) const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "Pidfile found: " << pid << ". Checking if process is still runnig..." << endl;
        log_notice(msg.str());
    }
}


/**
 * Another process is already running.
 * @param pid The pid of the other process.
 */
void Logger::print_process_already_running(const int pid) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Another bpdyndnsd process with PID " << pid << " is already running!" << endl;
        log_error(msg.str());
    }
}


/**
 * SIGTERM caught.
 */
void Logger::print_caught_sigterm() const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Caught SIGTERM. Exiting..." << endl;
        log_notice(msg.str());
    }
}


/**
 * SIGUSR1 caught.
 */
void Logger::print_caught_siguser1() const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Caught SIGUSR1. Switching to offline mode..." << endl;
        log_notice(msg.str());
    }
}


/**
 * SIGHUP caught.
 */
void Logger::print_caught_sighup() const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Caught SIGHUP. Reloading config and switching to online mode..." << endl;
        log_notice(msg.str());
    }
}


/**
 * Error while setting signal handler.
 */
void Logger::print_error_setting_signal(const string& signal) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Error while setting signal handler for: " << signal << endl;
        log_error(msg.str());
    }
}


/**
 * Error while setting signal handler.
 */
void Logger::print_init_log_facility() const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "Initialized logging facility." << " Loglevel: " << Loglevel << " Syslog: " << Syslog << endl;
        log_notice(msg.str());
    }
}


/**
 * Be verbose. Currently we are in offline mode.
 */
void Logger::print_offline_mode() const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Offline mode..." << endl;
        log_notice(msg.str());
    }
}


/**
 * Objects successfully serialized.
 */
void Logger::print_serialized_objects_success() const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "Serialized objects successfully." << endl;
        log_notice(msg.str());
    }
}


/**
 * Objects successfully de-serialized.
 */
void Logger::print_deserialized_objects_success() const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "De-serialized objects successfully." << endl;
        log_notice(msg.str());
    }
}


/**
 * Prints out the content of a service object.
 * @param message Message to be added on output first.
 * @param protocol Service's protocol.
 * @param hostname Service's hostname.
 * @param login Service's login.
 * @param password Service's password.
 * @param actual_ip Service's actual_ip.
 * @param lastupdated Service's lastupdated.
 */
void Logger::print_service_object(const string& message, const string& protocol, const string& hostname, const string& login, const string& password, const int update_interval, const int max_updates_within_interval, const string& actual_ip, list<int>* lastupdated) const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << message << endl;
        msg << "\t" << "Protocol:         " << protocol << endl;
        msg << "\t" << "Hostname:         " << hostname << endl;
        msg << "\t" << "Login:            " << login << endl;
        msg << "\t" << "Password:         " << password << endl;
        msg << "\t" << "Update Interval:  " << update_interval << endl;
        msg << "\t" << "Max Updates:      " << max_updates_within_interval << endl;
        msg << "\t" << "Actual_IP:        " << actual_ip << endl;
        BOOST_FOREACH( int update_time, *lastupdated)
        {
            msg << "\t" << "Lastupdated:      " << update_time << endl;
        }
        log_notice(msg.str());
    }
}


/**
 * Caught exception while serialize.
 * @param exception Exception message.
 */
void Logger::print_exception_serialize(const string& exception) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Error while trying to serialize Serviceholder object: " << exception << endl;
        log_error(msg.str());
    }
}


/**
 * Caught exception while de-serialize.
 * @param exception Exception message.
 */
void Logger::print_exception_deserialize(const std::string& exception) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Error while trying to de-serialize Serviceholder object: " << exception << endl;
        log_error(msg.str());
    }
}


/**
 * Child couldn't be killed by parent.
 * @param pid Pid of the child.
 */
void Logger::print_error_kill_child(const int pid) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Could not kill child process with PID: " << pid << endl;
        log_error(msg.str());
    }
}


/**
 * Child was killed by parent because of error.
 * @param pid The pid (child) killed.
 */
void Logger::print_child_killed(const int pid) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Killed child process with PID: " << pid << endl;
        log_notice(msg.str());
    }
}


/**
 * There is no object file.
 * @param object_file The object file.
 */
void Logger::print_no_object_file(const std::string& object_file) const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "There is no object file: " << object_file << ". Continue without recovering state from old services!" << endl;
        log_warning(msg.str());
    }
}


/**
 * Prints out the given hostname
 * @param hostname Hostname as string.
 */
void Logger::print_hostname(const std::string& hostname) const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "Detected following hostname of localhost: " << hostname << endl;
        log_notice(msg.str());
    }
}


/**
 * Prints out the detected own ipv4 address
 * @param ip_addr String representation of the detected ip.
 */
void Logger::print_own_ipv4(const std::string& ip_addr_v4) const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "Detected following IPv4-Address of this host: " << ip_addr_v4 << endl;
        log_notice(msg.str());
    }
}


/**
 * Prints out the detected own ipv5 address
 * @param ip_addr String representation of the detected ip.
 */
void Logger::print_own_ipv6(const std::string& ip_addr_v6) const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "Detected following IPv6-Address of this host: " << ip_addr_v6 << endl;
        log_notice(msg.str());
    }
}


/**
 * Exception while trying to resolve hostname to ip.
 * @param exception The exception caught.
 * @param hostname The hostname.
 */
void Logger::print_error_hostname_to_ip(const string& exception, const string& hostname) const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "Could not resolve the hostname: " << hostname << "to an IP-Address: " << exception << endl;
        log_error(msg.str());
    }
}


/**
 * The update of the given service was successful.
 * @param service The service.
 */
void Logger::print_update_service_successful(const string& service) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Updated service successful: " << service << endl;
        log_notice(msg.str());
    }
}


/**
 * No ip could be determined through webcheck
 */
void Logger::print_webcheck_no_ip() const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "This hosts' IP could not be determined through any configured webcheck url." << endl;
        log_error(msg.str());
    }
}


/**
 * Connection problem while trying to get ip through webcheck url
 * @param curl_err_buff Curl error message
 * @param url the url
 */
void Logger::print_webcheck_url_connection_problem(const char * curl_err_buff, const string& url) const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "There was a problem while trying to connect to following URL: " << url << " CURL error: " << curl_err_buff << endl;
        log_warning(msg.str());
    }
}


/**
 * Prints out curl error.
 * @param curl_err_buff Curl error message.
 * @param url URL
 */
void Logger::print_webcheck_error(const char * curl_err_buff, const string& url) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "There was an error while trying to connect to following URL: " << url << " CURL error: " << curl_err_buff << endl;
        log_error(msg.str());
    }
}


/**
 * Prints out the received data through curl.
 * @param curl_data Data string
 */
void Logger::print_received_curl_data(const string& curl_data) const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "Received CURL data: " << curl_data << endl;
        log_notice(msg.str());
    }
}


/**
 * IP was foudn through regex
 * @param ip The IP found.
 */
void Logger::print_regex_found_ip(const string& ip) const
{
    if ( 1 <= Loglevel )
    {
        ostringstream msg;
        msg << "Found IP-Address via regex: " << ip << endl;
        log_notice(msg.str());
    }
}


/**
 * No IP was found through regex.
 * @param data The data string which should contain a valid IP.s
 */
void Logger::print_regex_ip_not_found(const string& data) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Could not extract an IP-Address via regex from following data:\n" << data << endl;
        log_warning(msg.str());
    }
}


/**
 * Detected multiple occurrences of the same option.
 * @param message Error message.
 */
void Logger::print_multiple_cmd_option(const string& message) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "The same option is only allowed once: " << message << endl;
        log_error(msg.str());
    }
}


/**
 * An update would exceed the update interval. Prints out a warning message.
 * @param current_time Current time.
 * @param old_time Time of update #MaxUpdatesWithinInterval ago.
 * @param MaxUpdatesWithinInterval Number of allowed updates in one update interval.
 * @param service The service which exceeds update interval.
 */
void Logger::print_update_not_allowed(const int current_time, const int old_time, const int MaxUpdatesWithinInterval, const string& service) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Update not allowed for service: " << service << ". Too many updates within max update interval. Current time: " << current_time << ". Update time before " << MaxUpdatesWithinInterval << " updates: " << old_time << endl;
        log_warning(msg.str());
    }
}


/**
 * Failure while running update for service.
 * @param service Services' name.
 */
void Logger::print_update_service_failure(const std::string& service) const
{
    if ( 0 <= Loglevel )
    {
        ostringstream msg;
        msg << "Could not update service: " << service << endl;
        log_warning(msg.str());
    }
}
