/** @file
 * @brief The main function.
 *
 *
 *
 * @copyright Intra2net AG
 * @license GPLv2
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define VERSION     0
#define REVISION    1
#define RELEASE     0

#define PIDFILE "/var/run/bpdyndnsd.pid"
#define OBJECT_FILE "/home/bjoern/intranator/bpdyndnsd/objects.ser"

#include <iostream>
#include <fstream>
#include <list>
#include <string>
#include <boost/foreach.hpp>
#include <sys/types.h>
#include <signal.h>

#include "updater.h"

#include "config.cpp"
#include "dhs.cpp"
#include "logger.cpp"
#include "ods.cpp"
#include "service.cpp"
#include "serviceholder.cpp"
#include "updater.cpp"
#include "iphelper.cpp"


using namespace std;

Updater::Ptr updater;
bool is_online = true;

/**
 * Checks if a bpdyndnsd process is already running.
 * @param updater Shared Pointer to updater, needed for logging.
 * @return 0 if process not running already or pid of already running process.
 */
int check_for_running_process()
{
    ifstream pidfile(PIDFILE);
    if ( pidfile.is_open() )
    {
        int pid;
        pidfile >> pid;
        if ( pid > 0 )
            updater->get_logger()->print_pid_found(pid);
        else
            return 0;

        // check if process still running ret_val==-1 -> not runnig, ret_val==0 -> running
        if ( kill(pid,0) == 0)
        {
            updater->get_logger()->print_process_already_running(pid);
            pidfile.close();
            return pid;
        }
    }
    pidfile.close();
    return 0;
}


/**
 * Writes the pid into the pidfile.
 * @param pid The process's pid.
 */
int write_pidfile(int pid)
{
    ofstream pidfile(PIDFILE);
    if ( pidfile.is_open() )
    {
        pidfile << pid << endl;
    }
    else
    {
        updater->get_logger()->print_error_opening_rw(PIDFILE);
        return -1;
    }
    pidfile.close();
    return 0;
}


/**
 * Signal SIGTERM caught, releasing resources and exit.
 * @param param Parameter from the signal interface.
 */
void terminate(int param)
{
    updater->get_logger()->print_caught_sigterm();

    // unfortunately we can't call serialize_services in any destructor
    // cause some singleton are already destroyed !?!
    if ( updater->get_config()->serialize_services() != 0 )
    {
        updater.reset();
        exit(-1);
    }
    updater.reset();
    exit(0);
}



/**
 * Signal SIGUSR1 caught, switching to offline mode.
 * @param param Parameter from the signal interface.
 */
void switch_to_offline(int param)
{
    updater->get_logger()->print_caught_siguser1();
    is_online = false;
}


/**
 * Signal SIGHUP caught, reloading config and switching to online mode.
 * @param param Parameter from the signal interface.
 */
void reload_config(int param)
{
    updater->get_logger()->print_caught_sighup();
    if ( updater->reload_config() != 0 )
        exit(-1);
    is_online = true;
}


/**
 * Initialize the signals we handle.
 */
int init_signals()
{
    sighandler_t ret_val;
    ret_val = signal(SIGTERM,terminate);
    if ( ret_val == SIG_ERR )
    {
        updater->get_logger()->print_error_setting_signal("SIGTERM");
        return -1;
    }
    ret_val = signal(SIGUSR1,switch_to_offline);
    if ( ret_val == SIG_ERR )
    {
        updater->get_logger()->print_error_setting_signal("SIGUSR1");
        return -1;
    }
    ret_val = signal(SIGHUP,reload_config);
    if ( ret_val == SIG_ERR )
    {
        updater->get_logger()->print_error_setting_signal("SIGHUP");
        return -1;
    }

    return 0;
}


/**
 * Try to run in daemon mode if enabled in config.
 * @param daemon_mode True if process should detach to init (run as daemon), false if not detach to init.
 * @return 0 if all is fine, -1 on error.
 */
int init_daemon_mode(bool daemon_mode)
{
    updater->get_logger()->print_daemon_mode(daemon_mode);
    if ( daemon_mode == true )
    {
        int pid = fork();
        if ( pid < 0 )
        {
            // error fork
            updater->get_logger()->print_error_fork();
            return -1;
        }
        else if ( pid > 0 )
        {
            // parent continues here
            if ( write_pidfile(pid) != 0 )
            {
                if ( kill(pid,SIGTERM) != 0 )
                    updater->get_logger()->print_error_kill_child(pid);
                else
                    updater->get_logger()->print_child_killed(pid);
                exit(-1);
            }
            updater->get_logger()->print_runnig_as_daemon(pid);
            exit(0);
        }
        // child starts here
    }
    return 0;
}

/**
 * @brief The main part.
 * @param argc Number of arguments
 * @param argv Command line arguments
 * @return 0 if all is fine.
 */
int main(int argc, char *argv[])
{
    // initialize Updater
    Updater::Ptr _updater(new Updater);
    updater = _updater;
    _updater.reset();

    // load the cmd options
    if ( updater->init_config_from_cmd(argc,argv) != 0 )
        return -1;

    // load the config and service files
    if ( updater->init_config_from_files() != 0 )
        return -1;

    // open pidfile and check for running process
    if ( check_for_running_process() != 0)
        return -1;

    // init signal handling
    if ( init_signals() != 0)
        return -1;

    // init the ip_helper
    if ( updater->init_ip_helper() != 0 )
        return -1;

    // init daemon_mode if enabled
    if ( init_daemon_mode(updater->get_config()->get_daemon_mode()) != 0 )
        return -1;

    // service processing starts here
    do
    {
        if ( is_online == true )
        {
            // update all configured services
            updater->update_services();
        }
        else
        {
            // We are in offline mode, do nothing, expect printing "offline mode".
            updater->get_logger()->print_offline_mode();
        }

        // Snore, snore... don't hug the cpu if we are in daemon_mode.
        if ( updater->get_config()->get_daemon_mode() == 1 )
            sleep(5);

    }while ( updater->get_config()->get_daemon_mode() == 1 );


    // Serialize services to save their actual state.
    if ( updater->get_config()->serialize_services() != 0 )
        return -1;

    return 0;
}
