/** @file
 * @brief The updater class implementation. This class implements the updater logic.
 *
 *
 *
 * @copyright Intra2net AG
 * @license GPLv2
*/

#include "updater.h"

#include <boost/foreach.hpp>

using namespace std;

/**
 * Default constructor which initializes the member Conf.
 */
Updater::Updater()
    : IPHelp(new IPHelper)
{
    // Initialize program wide Logger, at this point we don't know where to log and with which loglevel.
    Logger::Ptr _log(new Logger);
    Log = _log;
    _log.reset();

    // initialize Config
    Config::Ptr _config(new Config(Log));
    Conf = _config;
    _config.reset();
}


/**
 * Default destructor.
 */
Updater::~Updater()
{
}


/**
 * Parse the command line arguments and initialize corresponding options.
 * @param argc Number command line arguments.
 * @param argv[] Array with arguments.
 * @return 0 if cmd options successfully parsed, 1 if usage or version.
 */
int Updater::init_config_from_cmd(int argc, char *argv[])
{
    // Load the command line parameters
    if( Conf->parse_cmd_line( argc, argv ) != 0)
        return -1;

    // If we have loaded the cmd options we need to init the log facility immediately in case debugging is enabled from cmd.
    init_log_facility();

    // successful parsed
    Log->print_cmd_parsed();
    return 0;
}


/**
 * Load the main config and the service definition files in config path.
 * @return 0 if all is fine, 
 */
int Updater::init_config_from_files()
{
    // Load the main and service config files in config path
    if ( Conf->load_config_from_files() != 0 )
        return -1;

    // Re-init log facility, perhaps new config file options for logger are set.
    // These config file options will only overwrite the cmd options if the SIGHUP (reload config) is caught.
    init_log_facility();

    // Here we are. All Service Objects should be initialized, so it is time to deserialize the old service objects and compare them.
    if ( Conf->deserialize_services() != 0 )
        return -1;

    // successful loaded
    Log->print_conf_files_parsed();
    return 0;
}


/**
 * Getter for member Config.
 * @return Member Config.
 */
Config::Ptr Updater::get_config() const
{
    return Conf;
}


/**
 * Getter for member Logger.
 * @return Member Logger.
 */
Logger::Ptr Updater::get_logger() const
{
    return Log;
}


/**
 * Init the IPHelp member with needed values.
 * @return 0 if all is fine, -1 otherwise.
 */
int Updater::init_ip_helper()
{
    // initialize IPHelper
    IPHelper::Ptr _iphelp(new IPHelper(Log,Conf->get_webcheck_ip_url(),Conf->get_webcheck_ip_url_alt(),Conf->get_enable_ipv6()));
    IPHelp = _iphelp;
    _iphelp.reset();

    return 0;
}


/**
 * Reloading the config. Delete all Service objects and then init new Service objects from config files.
 */
int Updater::reload_config()
{
    // serialize all service objects
    if ( Conf->serialize_services() != 0 )
        return -1;

    // delete all service objects
    Conf->delete_services();

    // delete the actual Variables_map, perhaps with old cmd options which would overwrite new config file options.
    Conf->delete_variables_map();

    // load only config files
    if ( init_config_from_files() != 0 )
        return -1;

    return 0;
}


/**
 * Initialize the logging facility with loglevel and syslog.
 */
void Updater::init_log_facility()
{
    Log->set_log_facility(Conf->get_loglevel(),Conf->get_syslog());
    Log->print_init_log_facility();
}


/**
 * Update all configured services.
 */
void Updater::update_services()
{
    list<Service::Ptr> services = this->Conf->get_services();

    string ip = IPHelp->get_actual_ip();

    if ( !ip.empty() )
    {
        BOOST_FOREACH(Service::Ptr &service, services )
        {
            service->update(ip);
        }
    }
}
